------------------------------------------------------------------------------
--
-- Filename:                subprogs.vhd
-- Inhalt:                  Beispiele fuer Funktionen und Prozeduren
--                          - Funktion: Konvertierung von bit in integer
--                          - Funktion: Zaehlen von Einsen in einem Bitvektor
--                          - Funktion: Exklusiv-NOR fuer Bitvektoren 
--                          - Prozedur: Ausgabe einer Meldung
--                          - Prozedur: D-Flip-Flop
--
------------------------------------------------------------------------------
--
-- Autoren:                 Bernhard Wunder, Gunther Lehmann
-- Letzte Aenderung:        04.03.94
--
------------------------------------------------------------------------------


PACKAGE my_pack IS

  FUNCTION bit_to_integer (bit_a : bit) RETURN integer ;
  FUNCTION count_ones (a : bit_vector)  RETURN integer ;
  FUNCTION exnor (a, b : bit_vector)    RETURN bit_vector ;
  PROCEDURE hello ;
  PROCEDURE d_ff (CONSTANT delay : IN  time := 2 ns ;
                  SIGNAL d,clk   : IN  bit ;
                  SIGNAL q,q_bar : OUT bit ) ;

END my_pack ;


------------------------------------------------------------------------------

PACKAGE BODY my_pack IS

  FUNCTION bit_to_integer (bit_a : bit) RETURN integer IS
  BEGIN
    IF bit_a = '1' THEN RETURN 1 ;
    ELSE RETURN 0 ;
    END IF ;
  END bit_to_integer ;


  FUNCTION count_ones (a : bit_vector) RETURN integer IS
    VARIABLE count : integer := 0 ;
  BEGIN
    FOR c IN a'RANGE LOOP
      IF a(c) = '1' THEN count := count + 1 ;
      END IF ;
    END LOOP ;
    RETURN count ;
  END count_ones ;


  FUNCTION exnor (a, b : bit_vector) RETURN bit_vector IS
    ALIAS c : bit_vector(1 TO a'LENGTH) IS a ;
    ALIAS d : bit_vector(1 TO b'LENGTH) IS b ;
    VARIABLE result : bit_vector(1 TO a'LENGTH) ;
  BEGIN
    ASSERT a'LENGTH = b'LENGTH REPORT "Different Length of vectors!" ;
    FOR k in c'RANGE LOOP
      IF (c(k) = '1' AND d(k) = '0') OR
         (c(k) = '0' AND d(k) = '1') THEN result(k) := '0' ;
      ELSE result(k) := '1' ;
      END IF ;
    END LOOP ;
    RETURN result ;
  END exnor ;


  PROCEDURE hello IS
  BEGIN
    ASSERT false REPORT "Hello world!" SEVERITY note ;
  END hello ;


  PROCEDURE d_ff (CONSTANT delay : IN  time := 2 ns ;
                  SIGNAL d,clk   : IN  bit ;
                  SIGNAL q,q_bar : OUT bit ) IS
  BEGIN
    IF clk = '1' AND clk'EVENT THEN 
      q     <= d     AFTER delay ;
      q_bar <= NOT d AFTER delay ;
    END IF ;
  END d_ff ;


END my_pack ;



------------------------------------------------------------------------------
-- Anwendung der Prozedur d_ff: N-Bit-Register aus D-Flip-Flops

ENTITY four_bit_register IS

  PORT (clk              : IN  bit ;
        in_d             : IN  bit_vector(3 DOWNTO 0) ;
        out_q, out_q_bar : OUT bit_vector(3 DOWNTO 0) ) ;
   
END four_bit_register ;

------------------------------------------------------------------------------

ARCHITECTURE with_procedures OF four_bit_register IS

  USE work.my_pack.ALL ;

BEGIN

  -- nebenlaeufige Prozeduraufrufe

  d_ff_3 : d_ff (1.5 ns, in_d(3), clk, 
                 out_q(3), out_q_bar(3)) ;
  d_ff_2 : d_ff (delay => 1.7 ns, d => in_d(2), clk => clk, 
                 q => out_q(2), q_bar => out_q_bar(2)) ;
  d_ff_1 : d_ff (1.9 ns, in_d(1), clk, 
                 q_bar => out_q_bar(1), q => out_q(1)) ;
  d_ff_0 : d_ff (q_bar => out_q_bar(0), clk => clk,
                 q => out_q(0), d => in_d(0)) ;

END with_procedures ;



------------------------------------------------------------------------------
-- Anwendung der Funktionen exnor, bit_to_integer, count_ones und der
-- Prozedur hello

USE work.my_pack.ALL ;
ENTITY functions IS
  
  PORT (in1, in2  : IN  bit_vector(8 DOWNTO 1) := "11111111";
        exnor_out : OUT bit_vector(8 DOWNTO 1));

END functions;


------------------------------------------------------------------------------

ARCHITECTURE behavioral OF functions IS

  SIGNAL x2:   bit_vector(2 DOWNTO 1);
  SIGNAL i, j: integer := 0;

BEGIN
  
  exnor_out <= exnor(in1, in2);                        -- Vektoren Laenge 8

  x2 <= exnor(a => in1(2 DOWNTO 1), b => in2(2 DOWNTO 1));      -- Laenge 2

  i <= bit_to_integer(bit_a => in1(8)) + count_ones(in2);

  j <= count_ones("11101001011100101001001");               -- Ergebnis: 12

  hello;                         -- einmaliger Aufruf bei Simulationsbeginn

END behavioral;

------------------------------------------------------------------------------






  

